from nanomotif.candidate import *
from nanomotif.constants import *
import pytest
from hypothesis import given, strategies as st
import itertools
import pandas as pd
import polars.testing

# Strategy for a single base
base = st.sampled_from("ATCG.")
# Strategy for a bracketed sequence
bracketed = st.text(min_size=2, max_size=3, alphabet="ATCG", ).map(lambda s: f"[{''.join(list(set(s)))}]")
# Strategy for a list with bracketed and unbracketed sequences
combined = st.lists(st.one_of(base, bracketed), min_size=1)

class TestMotif:
    @given(st.text(alphabet=["A", "C", "G", "T", "."], min_size=5, max_size=10))
    def test_strip(self, motif_string):
        motif = Motif(motif_string, 0)
        stripped = motif.strip()
        assert stripped == motif_string.lstrip(".").rstrip(".")
    
    @given(combined)
    def test_split(self, string_list):
        motif_string = "".join(string_list)
        motif = Motif(motif_string, 0)
        split_motif = motif.split()
        # Splitting of motif removes braces
        assert split_motif == string_list

    @given(combined)
    def test_array(self, motif_list):
        motif_string = "".join(motif_list)
        motif = Motif(motif_string, 0)
        array = motif.one_hot()
        assert array.shape == (len(motif_list), 4)
        assert array.dtype == int
        assert array.min() >= 0
        assert array.max() <= 1
    
    def test_reverse_compliment(self):
        motif = Motif("ATCG", 0)
        assert motif.reverse_compliment() == Motif("CGAT", 3)
        motif = Motif("AT[CG]G", 0)
        assert motif.reverse_compliment() == Motif("C[CG]AT", 3)
        motif = Motif("ATC.G.", 0)
        assert motif.reverse_compliment() == Motif(".C.GAT", 3)
        motif = Motif("ATAC.G.", 2)
        assert motif.reverse_compliment() == Motif(".C.GTAT", 4)
    
    def test_new_stripped_motif(self):
        motif = Motif("....ATCG", 4)
        assert motif.new_stripped_motif() == Motif("ATCG", 0)
        motif = Motif("ATCG....", 0)
        assert motif.new_stripped_motif() == Motif("ATCG", 0)
        motif = Motif("....AT..CG", 4)
        assert motif.new_stripped_motif() == Motif("AT..CG", 0)
    
    def test_strip_reverse(self):
        motif = Motif("....AT..CG..", 4)
        motif_strip = motif.new_stripped_motif()
        motif_strip_reverse = motif_strip.reverse_compliment()
        assert motif_strip_reverse == Motif("CG..AT", 5)
    
    def test_sub_motif_of(self):
        motif1 = Motif("ATCG", 2)
        motif2 = Motif("ATCG", 2)
        assert motif1.sub_motif_of(motif2) == True

        motif1 = Motif("ATCG", 0)
        motif2 = Motif("AT", 0)
        assert motif1.sub_motif_of(motif2) == True

        motif1 = Motif("A[TCG]CG", 0)
        motif2 = Motif("A.C", 0)
        assert motif1.sub_motif_of(motif2) == True

        motif1 = Motif("ATCG", 0)
        motif2 = Motif("AT", 0)
        assert motif1.sub_motif_of(motif2) == True

        motif1 = Motif("ATCG", 2)
        motif2 = Motif("CG", 0)
        assert motif1.sub_motif_of(motif2) == True

        motif1 = Motif("ATCG", 2)
        motif2 = Motif("CG", 2)
        assert motif1.sub_motif_of(motif2) == False

        motif1 = Motif("CG", 1)
        motif2 = Motif("ATCG", 2)
        assert motif1.sub_motif_of(motif2) == False

    def test_sub_string_of(self):
        # Equal motifs strings
        motif1 = Motif("ATCG", 2)
        motif2 = Motif("ATCG", 0)
        assert motif1.sub_string_of(motif2) == True

        # Single base difference
        motif1 = Motif("AGCG", 2)
        motif2 = Motif("ATCG", 0)
        assert motif1.sub_string_of(motif2) == False

        # Longer sub string of shorter
        motif1 = Motif("ATCG", 0)
        motif2 = Motif("AT", 0)
        assert motif1.sub_string_of(motif2) == True

        # Workrking with braces
        motif1 = Motif("ACC", 0)
        motif2 = Motif("A[CG]C", 2)
        assert motif1.sub_string_of(motif2) == True

        # Working with dots
        motif1 = Motif("ATCG", 0)
        motif2 = Motif("AT.G", 2)
        assert motif1.sub_string_of(motif2) == True

        # Works when braces in motif1
        motif1 = Motif("AT.G", 0)
        motif2 = Motif("ATCG", 2)
        assert motif1.sub_string_of(motif2) == False

        # Works when braces in motif1
        motif1 = Motif("AT[ACG]G", 0)
        motif2 = Motif("ATCG", 0)
        assert motif1.sub_string_of(motif2) == False

        # Works when motif1 is longer
        motif1 = Motif("T.....ATCG...C", 0)
        motif2 = Motif("ATCG", 2)
        assert motif1.sub_string_of(motif2) == True

        # Works when motif2 is longer
        motif1 = Motif("ATCG", 2)
        motif2 = Motif("T.....ATCG...C", 0)
        assert motif1.sub_string_of(motif2) == False

        # Dots and braces in both
        motif1 = Motif("[AT]..AT..CG[TA]C..C", 2)
        motif2 = Motif("[TA]T..CG[AT][GC]", 0)
        assert motif1.sub_string_of(motif2) == True

    def test_distance(self):
        motif1 = Motif("A[TCG]CG", 0)
        motif2 = Motif("A.C", 0)
        assert motif1.distance(motif2) == 2

        motif1 = Motif("ATCG", 0)
        motif2 = Motif("AT", 0)
        assert motif1.distance(motif2) == 2

        motif1 = Motif("ATCG", 2)
        motif2 = Motif("CG", 0)
        assert motif1.distance(motif2) == 2

        motif1 = Motif("CG", 0)
        motif2 = Motif("ATCG", 2)
        assert motif1.distance(motif2) == 2

        motif1 = Motif("C...ATCG", 6)
        motif2 = Motif("CG", 0)
        assert motif1.distance(motif2) == 3

        motif1 = Motif("CG", 0)
        motif2 = Motif("C...ATCG", 6)
        assert motif1.distance(motif2) == 3

        motif1 = Motif("C..CG...G", 3)
        motif2 = Motif("CG", 0)
        assert motif1.distance(motif2) == 2

        motif1 = Motif("CG", 0)
        motif2 = Motif("C..CG...G", 3)
        assert motif1.distance(motif2) == 2

        motif1 = Motif("CG", 0)
        motif2 = Motif("C..CG...[GC]", 3)
        assert motif1.distance(motif2) == 2
    def test_merge(self):
        motif1 = Motif("..A..", 2)
        motif2 = Motif("..C..", 2)
        assert motif1.merge(motif2) == Motif("..[AC]..", 2)

        motif1 = Motif(".ATGC.", 1)
        motif2 = Motif(".TCAG", 2)
        merged = motif1.merge(motif2)
        assert merged == Motif("..[AC][AT]G..", 2)
        assert merged.mod_position == 2

        motif1 = Motif(".[AT]T.C.[CT]T", 2)
        motif2 = Motif(".[GC].CA.[CG]G", 2)
        merged = motif1.merge(motif2)
        assert merged == Motif("....[AC].[CGT][GT]", 2)
        assert merged.mod_position == 2
    
    def test_iupac(self):
        motif = Motif("A[TCG]CG", 0)
        assert motif.iupac() == "ABCG"
        motif = Motif("A.C", 0)
        assert motif.iupac() == "ANC"
        motif = Motif("ATCG", 0)
        assert motif.iupac() == "ATCG"
        motif = Motif("ATCG", 2)
        assert motif.iupac() == "ATCG"
        motif = Motif("CG", 0)
        assert motif.iupac() == "CG"
        motif = Motif("C...ATCG", 6)
        assert motif.iupac() == "CNNNATCG"
        motif = Motif("C..CG...G", 3)
        assert motif.iupac() == "CNNCGNNNG"
        motif = Motif("C..CG...[GC]", 3)
        assert motif.iupac() == "CNNCGNNNS"

        
        






import numpy as np
import polars as pl
import os
from polars import col
import logging as log
import itertools
from itertools import takewhile
from scipy.stats import entropy
import multiprocessing
from multiprocessing import get_context
import progressbar
from nanomotif.constants import *
from nanomotif.model import BetaBernoulliModel
from nanomotif.utils import subseq_indices, calculate_match_length
from nanomotif.seq import EqualLengthDNASet, DNAsequence
from nanomotif.candidate import Motif, MotifTree
from nanomotif.logger import configure_logger
from nanomotif.parallel import update_progress_bar
import heapq as hq
import networkx as nx
import warnings
import time


##########################################
# Motif candidate scoring
##########################################

def methylated_motif_occourances(motif: str, sequence: str, methylated_positions) -> tuple:
    """
    Get occourances of a motif in a contig

    Parameters:
    - motif (str): The motif to search for.
    - seq (str): The contig to search in.
    - contig_meth_positions (list): The positions of the methylation sites in the contig.

    Returns:
    - tuple: A tuple of two numpy arrays, the first containing the positions of the methylated motifs, the second containing the positions of the non-methylated motifs.
    """


    # Get the index of the methylation in the motif in the contig
    motif_index = subseq_indices(motif.string, sequence) + motif.mod_position

    # Methylated motif positions
    meth_occurences = np.intersect1d(methylated_positions, motif_index)

    # Non-methylated motif positions
    nonmeth_occurences =  np.setdiff1d(motif_index, methylated_positions)

    return meth_occurences, nonmeth_occurences

def motif_model_contig(pileup, contig: str, motif):
    """
    Get the posterior for a single motif. Uses number of methylated motifs as methylation count.

    Parameters:
    - pileup (Pileup): The pileup to be processed.
    - contig (str): The contig to be processed.
    - motif (str): The motif to be processed.
    - motif_meth_index (int): The index of the methylation site in the motif.

    Returns:
    - BetaBernoulliModel: The model for the candidate.
    """
    # Strip motif of periods    # Strip motif of periods
    motif_stripped = motif.new_stripped_motif()

    meth_positions_fwd = pileup.filter(pl.col("strand") == "+")["position"].to_numpy()
    meth_positions_rev = pileup.filter(pl.col("strand") == "-")["position"].to_numpy()

    index_meth_fwd, index_nonmeth_fwd = methylated_motif_occourances(motif_stripped, contig, meth_positions_fwd)
    index_meth_rev, index_nonmeth_rev = methylated_motif_occourances(motif_stripped.reverse_compliment(), contig, meth_positions_rev)

    model = BetaBernoulliModel()
    model.update(len(index_meth_fwd) + len(index_meth_rev), len(index_nonmeth_fwd) + len(index_nonmeth_rev))

    return model


def methylated_reads_counts(pileup: list, sequence: str, motif: str) -> tuple:
    motif_index = subseq_indices(motif.string, sequence) + motif.mod_position

    pileup = pileup.filter(pl.col("position").is_in(motif_index))
    counts = pileup \
        .select(["n_mod", "n_nonmod"]) \
        .sum()
    n_mod = counts.get_column("n_mod")[0]
    n_nonmod = counts.get_column("n_nonmod")[0]
    return n_mod, n_nonmod


def motif_model_read(pileup, contig: str, motif):
    """
    Get the posterior for a single motif. Uses number of methylated reads as methylation count.

    Parameters:
    - pileup (Pileup): The pileup to be processed.
    - read (str): The read to be processed.
    - motif (str): The motif to be processed.

    Returns:
    - BetaBernoulliModel: The model for the candidate.
    """
    motif_stripped = motif.new_stripped_motif()
    pileup_counts = pileup.with_columns((pl.col("fraction_mod")*pl.col("Nvalid_cov")).alias("n_mod")) \
        .with_columns(pl.col("n_mod").round().cast(pl.Int32)) \
        .with_columns((pl.col("Nvalid_cov") - pl.col("n_mod")).alias("n_nonmod"))
    
    pileup_plus = pileup_counts.filter(pl.col("strand") == "+")
    pileup_minus = pileup_counts.filter(pl.col("strand") == "-")

    n_meth_plus, n_nonmeth_plus = methylated_reads_counts(pileup_plus, contig, motif_stripped)
    n_meth_minus, n_nonmeth_minus = methylated_reads_counts(pileup_minus, contig, motif_stripped.reverse_compliment())

    model = BetaBernoulliModel()
    model.update(n_meth_plus + n_meth_minus, n_nonmeth_plus + n_nonmeth_minus)
    return model


##########################################
# Motif candidate state space search
##########################################




#
#def process_sample(assembly, pileup, 
#                   max_candidate_size = 40,
#                   min_read_methylation_fraction = 0.8,
#                   min_valid_coverage = 1,
#                   min_kl_divergence = 0.1
#                   ):
#    """
#    Process a single sample
#    
#    Parameters:
#    - assembly (Assembly): The assembly to be processed.
#    - pileup (Pileup): The pileup to be processed.
#    - max_candidate_size (int): The maximum size of the candidate motifs.
#    - min_read_methylation_fraction (float): The minimum fraction of reads that must be methylated for a position to be considered methylated.
#    - min_valid_coverage (int): The minimum number of reads that must cover a position for it to be considered valid.
#    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
#    - min_cdf_score (float): Minimum score of 1 - cdf(cdf_position) for a motif to be considered valid.
#    - cdf_position (float): The position to evaluate the cdf at.
#    - min_motif_frequency (int): Used to get minimum number of sequences to evaluate motif at.
#    """
#    assert pileup is not None, "Pileup is None"
#    assert len(pileup) > 0, "Pileup is empty" 
#    assert assembly is not None, "Assembly is None"
#    assert max_candidate_size > 0, "max_candidate_size must be greater than 0"
#    assert min_read_methylation_fraction >= 0 and min_read_methylation_fraction <= 1, "min_read_methylation_fraction must be between 0 and 1"
#    assert min_valid_coverage >= 0, "min_valid_coverage must be greater than 0"
#    assert min_kl_divergence >= 0, "min_kl_divergence must be greater than 0"
#
#    padding = max_candidate_size // 2
#    result = []
#    pileup = pileup \
#            .filter(pl.col("Nvalid_cov") > min_valid_coverage) \
#            .filter(pl.col("fraction_mod") >= min_read_methylation_fraction) \
#            .sort("contig") 
#    total_tasks = pileup.select(pl.struct(["contig", "mod_type"]).n_unique()).item()
#
#    #
#    widgets=[
#        progressbar.Timer(), ' | ',progressbar.Counter(), ' of ', str(total_tasks),' - ', progressbar.Percentage(), ' ',progressbar.Bar(),' (', progressbar.ETA(), ') '
#    ]
#    for (contig, modtype), subpileup in progressbar.progressbar(pileup.groupby(["contig", "mod_type"]), widgets=widgets, max_value=total_tasks, redirect_stdout=True):
#        log.info(f"Processing {contig} {modtype}")
#
#        contig_sequence = assembly.assembly[contig]
#        index_plus = subpileup.filter(pl.col("strand") == "+").get_column("position").to_list()
#        index_minus = subpileup.filter(pl.col("strand") == "-").get_column("position").to_list()
#        if len(index_minus) <= 1 or len(index_plus) <= 1:
#            log.info("Too few methylated positions")
#            continue
#
#        sequences_plus = contig_sequence.sample_at_indices(index_plus, padding)
#        sequences_minus = contig_sequence.sample_at_indices(index_minus, padding).reverse_compliment()
#        sequences = sequences_plus + sequences_minus
#        sequences_array = sequences.convert_to_DNAarray()
#
#        motif_graph, best_candidates = find_best_candidates(
#            sequences_array, 
#            contig_sequence, 
#            subpileup, 
#            min_kl = min_kl_divergence
#        )
#        identified_motifs = nxgraph_to_dataframe(motif_graph) \
#            .filter(col("sequence").is_in(best_candidates))
#        
#        if len(identified_motifs) == 0:
#            log.info("No motifs found")
#            continue
#        else:
#            result.append(identified_motifs.with_columns(
#                pl.lit(contig).alias("contig"),
#                pl.lit(modtype).alias("mod_type")
#            ))
#    if len(result) == 0:
#        return None
#    motifs = pl.concat(result) \
#        .with_columns([
#            pl.col("sequence").apply(lambda motif: motif[count_periods_at_start(motif):len(motif)-count_periods_at_end(motif)]).alias("motif"),
#            pl.col("sequence").apply(lambda motif: padding - count_periods_at_start(motif)).alias("mod_position")
#        ])
#    return motifs
#
#

def worker_function(args, counter, lock, assembly, min_kl_divergence, padding, minimum_methylation_fraction_confident, log_dir, verbose):
    """
    Process a single subpileup for one contig and one modtype

    Parameters:
    - args (tuple): The arguments to the function: contig, modtype, subpileup
    - counter (multiprocessing.Value): The progress counter
    - lock (multiprocessing.Lock): The lock for the progress counter
    - assembly (Assembly): The assembly to be processed.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - padding (int): The padding to use for the motif.
    """
    
    warnings.filterwarnings("ignore")
    contig, modtype, subpileup = args
    
    process_id = os.getpid()
    log_file = f"{log_dir}/find-motifs.{process_id}.log"
    configure_logger(log_file, verbose=verbose)

    try:
        result = process_subpileup(contig, modtype, subpileup, assembly, min_kl_divergence, padding, minimum_methylation_fraction_confident)
        with lock:
            counter.value += 1
        return result
    except:
        with lock:
            counter.value += 1
        return None





def process_subpileup(contig, modtype, subpileup, assembly, min_kl_divergence, padding, minimum_methylation_fraction_confident):
    """
    Process a single subpileup for one contig and one modtype

    Parameters:
    - args (tuple): The arguments to the function: contig, modtype, subpileup
    - counter (multiprocessing.Value): The progress counter
    - lock (multiprocessing.Lock): The lock for the progress counter
    - assembly (Assembly): The assembly to be processed.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - padding (int): The padding to use for the motif.
    """
    log.info(f"Processing {contig} {modtype}")

    contig_sequence = assembly.assembly[contig]
    subpileup_high_confident = subpileup.filter(pl.col("fraction_mod") >= minimum_methylation_fraction_confident)
    index_plus = subpileup_high_confident.filter(pl.col("strand") == "+").get_column("position").to_list()
    index_minus = subpileup_high_confident.filter(pl.col("strand") == "-").get_column("position").to_list()
    if len(index_minus) <= 1 or len(index_plus) <= 1:
        log.info("Too few methylated positions")
        return None

    sequences_plus = contig_sequence.sample_at_indices(index_plus, padding)
    sequences_minus = contig_sequence.sample_at_indices(index_minus, padding).reverse_compliment()
    sequences = sequences_plus + sequences_minus
    sequences_array = sequences.convert_to_DNAarray()

    motif_graph, best_candidates = find_best_candidates(
        sequences_array, 
        contig_sequence, 
        subpileup, 
        min_kl = min_kl_divergence
    )
    identified_motifs = nxgraph_to_dataframe(motif_graph) \
        .filter(col("sequence").is_in(best_candidates))
    
    if len(identified_motifs) == 0:
        log.info("No motifs found")
        return None
    else:
        identified_motifs = identified_motifs.with_columns(
            pl.lit(contig).alias("contig"),
            pl.lit(modtype).alias("mod_type")
        ).with_columns([
            pl.col("model").apply(lambda x: x._alpha).alias("alpha"),
            pl.col("model").apply(lambda x: x._beta).alias("beta")
        ]).drop("model")
        return identified_motifs
    

def process_sample_parallel(
        assembly, pileup, 
        threads = 2,
        search_frame_size = 40,
        threshold_methylation_confident = 0.8,
        threshold_methylation_general = 0.5,
        threshold_valid_coverage = 1,
        minimum_kl_divergence = 0.2,
        verbose = False,
        log_dir = None
    ):
    """
    Process a single sample
    
    Parameters:
    - assembly (Assembly): The assembly to be processed.
    - pileup (Pileup): The pileup to be processed.
    - max_candidate_size (int): The maximum size of the candidate motifs.
    - min_read_methylation_fraction (float): The minimum fraction of reads that must be methylated for a position to be considered methylated.
    - threshold_valid_coverage (int): The minimum number of reads that must cover a position for it to be considered valid.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - min_cdf_score (float): Minimum score of 1 - cdf(cdf_position) for a motif to be considered valid.
    - cdf_position (float): The position to evaluate the cdf at.
    - min_motif_frequency (int): Used to get minimum number of sequences to evaluate motif at.
    """
    assert pileup is not None, "Pileup is None"
    assert len(pileup) > 0, "Pileup is empty" 
    assert assembly is not None, "Assembly is None"
    assert search_frame_size > 0, "search_frame_size must be greater than 0"
    assert threshold_methylation_confident >= 0 and threshold_methylation_confident <= 1, "threshold_methylation_confident must be between 0 and 1"
    assert threshold_methylation_general >= 0 and threshold_methylation_general <= 1, "threshold_methylation_general must be between 0 and 1"
    assert threshold_valid_coverage >= 0, "min_valid_coverage must be greater than 0"
    assert minimum_kl_divergence >= 0, "mininum_kl_divergence must be greater than 0"

    # Infer padding size from candidate_size
    padding = search_frame_size // 2

    # Filter pileup
    pileup = pileup \
            .filter(pl.col("Nvalid_cov") > threshold_valid_coverage) \
            .filter(pl.col("fraction_mod") >= threshold_methylation_general) \
            .sort("contig") 
    
    # Create a list of tasks (TODO: not have a list of all data)
    tasks = [(contig, modtype, subpileup) for (contig, modtype), subpileup in pileup.groupby(["contig", "mod_type"])]

    # Create a progress manager
    manager = multiprocessing.Manager()
    counter = manager.Value('i', 0)
    lock = manager.Lock()

    # Create a pool of workers
    pool = get_context("spawn").Pool(processes=threads)

    # Create a process for the progress bar
    progress_bar_process = multiprocessing.Process(target=update_progress_bar, args=(counter, len(tasks)))
    progress_bar_process.start()

    # Put them workers to work
    results = pool.starmap(worker_function, [(
        task, 
        counter, lock, 
        assembly, minimum_kl_divergence, padding, threshold_methylation_confident, 
        log_dir, verbose
        ) for task in tasks])
    results = [result for result in results if result is not None]

    # Close the pool
    pool.close()
    pool.join()

    # Close the progress bar
    progress_bar_process.join()

    if len(results) == 0:
        return None
    motifs = pl.concat(results)

    model_col = []
    for a, b in zip(motifs.get_column("alpha").to_list(), motifs.get_column("beta").to_list()):
        model_col.append(BetaBernoulliModel(a, b))

    motifs = motifs.with_columns([
            pl.lit(model_col).alias("model"),
            pl.col("sequence").apply(lambda motif: motif[count_periods_at_start(motif):len(motif)-count_periods_at_end(motif)]).alias("motif"),
            pl.col("sequence").apply(lambda motif: padding - count_periods_at_start(motif)).alias("mod_position")
        ]).drop(["alpha", "beta"])

    return motifs




#########################################################################
# Motif candidate state space 

def find_best_candidates(methylation_sequences, sequence, pileup, min_kl = 0.2, max_dead_ends = 25):
    """
    Find the best motif candidates in a sequence.

    Parameters:
    - methylation_sequences (DNAarray): The methylation sequences extracted from the contig at methylated sites.
    - sequence (DNAsequence): The contig to be processed.
    - pileup (Pileup): The pileup to be processed.
    - min_kl (float): The minimum KL-divergence for a position in the candidate to be expanded
    - max_dead_ends (int): The maximum number of low scoring candidates before the search is terminated.
    """
    padding = methylation_sequences.shape[1] // 2
    total_sequences = methylation_sequences.shape[0]
    mod_type = pileup.get_column("mod_type").unique().to_list()[0]
    
    root_motif = Motif("." * padding + MOD_TYPE_TO_CANONICAL[mod_type] + "." * padding, padding) # Represent all possible motifs
    methylation_sequences_subset = methylation_sequences.copy()
    best_candidates = []
    continue_search = True
    dead_ends = 0
    motif_graph = None

    while continue_search:
        if dead_ends >= max_dead_ends:
            log.debug("Stopping search, too many low scoring candidates")
            break
        # Find the initial guess within the tree
        motif_graph, naive_guess = a_star_search(
            root_motif, 
            sequence, 
            pileup, 
            methylation_sequences_subset, 
            motif_graph = motif_graph,
            min_kl = min_kl,
            max_rounds_since_new_best = 15
        )

        # If there is no naive guess, stop the search
        if naive_guess == root_motif:
            log.debug("No naive guess found, stopping search")
            break
        guess = naive_guess
        # next_guess = naive_guess
        # while guess != next_guess:
        #     # Find the best guess within the subtree of the naive guess
        #     guess = next_guess
        #     motif_graph, next_guess = a_star_search(
        #         guess, 
        #         sequence, 
        #         pileup, 
        #         methylation_sequences_subset, 
        #         motif_graph = motif_graph,
        #         min_kl = min_kl,
        #         max_rounds_since_new_best = 5
        #     )

        # Remove new candidate from methylation sequences
        seq_before = methylation_sequences_subset.shape[0]
        methylation_sequences_subset = methylation_sequences_subset.filter_sequence_matches(naive_guess.one_hot(), keep_matches = False)
        if methylation_sequences_subset is None:
            log.debug("No more sequences left")
            break

        # Check if we should continue the search
        seq_remaining = methylation_sequences_subset.shape[0]
        seq_remaining_percent = seq_remaining/total_sequences
        log.debug(f"{naive_guess}, {seq_before-seq_remaining} seqs. model: {motif_graph.nodes[naive_guess]['model']}. ({100*seq_remaining_percent:.1f} % left)")

        if motif_graph.nodes[naive_guess]["score"] < 0.1:
            dead_ends += 1
            log.debug(f"Candidate has low score. {dead_ends} of {max_dead_ends} before temination")
            continue

        best_candidates.append(naive_guess)
        
        if (seq_remaining/total_sequences) < 0.01:
            log.debug("Stopping search, too few sequences left")
            break
        log.debug("Continuing search")
    return motif_graph, best_candidates


 
def motif_child_nodes_kl_dist_max(motif, meth_pssm, contig_pssm, freq_threshold=0.25, min_kl=0.1):
        kl_divergence = entropy(meth_pssm, contig_pssm)
        split_motif = motif.split()

        evaluated = np.array([i for i, base in enumerate(split_motif) if base != "."])
        kl_divergence[evaluated] = 0
        if np.max(kl_divergence) < min_kl:
            return
        
        pos = np.where(kl_divergence == np.max(kl_divergence))[0][0]

        # Methylation frequency most be above contig frequency
        index_meth_frequncies_highest = meth_pssm[:, pos] > contig_pssm[:, pos]

        # Methylation frequency most be above a threshold
        index_meth_frequncies_above_threshold = meth_pssm[:, pos] > freq_threshold

        # Combine the two filters
        index_position_filt = np.logical_and(index_meth_frequncies_highest, index_meth_frequncies_above_threshold)
        bases_index = np.argwhere(index_position_filt).reshape(-1)
        bases_filt = [BASES[int(i)] for i in bases_index]
        
        # All combination of the bases
        combinations = []
        for i in range(1, min(len(bases_filt)+1, 4)):
            combinations += list(itertools.combinations(bases_filt, i))
        
        for base in combinations:
            if len(base) > 1:
                base = "[" + "".join(list(base)) + "]"
            else:
                base = "".join(list(base))
            new_motif = split_motif[:pos] + [base] + split_motif[pos+1:]
            yield Motif("".join(new_motif), motif.mod_position)

 
def motif_child_nodes_kl_dist_prune(motif, meth_pssm, contig_pssm, min_kl=0.1, freq_threshold=0.35):
        kl_divergence = entropy(meth_pssm, contig_pssm)
        split_motif = motif.split()
        
        for pos in np.where(kl_divergence > min_kl)[0]:
            if split_motif[pos] != ".":
                continue

            # Methylation frequency most be above contig frequency
            index_meth_frequncies_highest = meth_pssm[:, pos] > contig_pssm[:, pos]

            # Methylation frequency most be above a threshold
            index_meth_frequncies_above_threshold = meth_pssm[:, pos] > freq_threshold

            # Combine the two filters
            index_position_filt = np.logical_and(index_meth_frequncies_highest, index_meth_frequncies_above_threshold)
            bases_index = np.argwhere(index_position_filt).reshape(-1)
            bases_filt = [BASES[int(i)] for i in bases_index]
            
            # All combination of the bases
            combinations = []
            for i in range(1, min(len(bases_filt)+1, 4)):
                combinations += list(itertools.combinations(bases_filt, i))
            
            for base in combinations:
                if len(base) > 1:
                    base = "[" + "".join(list(base)) + "]"
                else:
                    base = "".join(list(base))
                new_motif = split_motif[:pos] + [base] + split_motif[pos+1:]
                yield Motif("".join(new_motif), motif.mod_position)

def a_star_search(root_motif, contig, pileup, methylation_sequences, 
                  motif_graph = None, min_kl = 0.1, max_rounds_since_new_best = 5):
    """
    A* search for methylation motifs

    Parameters:
    - root_motif (list): The root motif to start the search from.
    - contig (str): The contig to be processed.
    - pileup (Pileup): The pileup to be processed.
    - methylation_sequences (DNAarray): The methylation sequences to be processed.
    """
    
    # Function for calculating the priority of a motif
    def priority_function(next_model, root_model):
        try:
            d_alpha = 1 - (next_model._alpha / root_model._alpha)
        except ZeroDivisionError:
            d_alpha = 1
        try:
            d_beta = next_model._beta / root_model._beta
        except ZeroDivisionError:
            d_beta = 1
        priority = d_alpha * d_beta
        return priority


    # Function for scoring a motif
    def scoring_function(next_model, current_model):
        mean_diff =  next_model.mean() - current_model.mean()
        return (next_model.mean() * -np.log10(next_model.standard_deviation())) * mean_diff

    # Search setup
    padding = methylation_sequences.shape[1] // 2
    total_sequences = methylation_sequences.shape[0]
    best_guess = root_motif
    root_model = motif_model_contig(pileup, contig.sequence, root_motif)
    best_score = scoring_function(root_model, root_model)
    rounds_since_new_best = 0
    visisted_nodes = []

    # Sample sequence in contig to get background for KL-divergence
    contig_sequences = contig.sample_n_subsequences(padding*2 + 1, 10000)
    contig_pssm = contig_sequences.pssm()

    # Initialize the search tree
    if motif_graph is None:
        motif_graph = MotifTree()
    motif_graph.add_node(root_motif, model=root_model, motif=root_motif, visited=False, score=scoring_function(root_model, root_model))

    # Initialize priority que
    priority_que = []
    hq.heappush(priority_que, (0, root_motif))

    # Search loop
    while len(priority_que) > 0:
        # Get the current best candidate
        current = hq.heappop(priority_que)[1]
        if current in visisted_nodes:
            continue
        if len(current.strip()) > 18:
            log.debug(f"{current}, Skipping due to length")
            continue


        current_model = motif_graph.nodes[current]["model"]
        visisted_nodes.append(current)
        log.debug(f"{''.join(current)} | {current_model} | {motif_graph.nodes[current]['score']:.2f} | {len(priority_que)}")
        motif_graph.nodes[current]["visited"] = True
        rounds_since_new_best += 1
        active_methylation_sequences = methylation_sequences.copy().filter_sequence_matches(current.one_hot(), keep_matches = True)
        if active_methylation_sequences is None:
            log.debug("No more sequences left")
            continue
        # Prune the neibhors of the current candidate
        neighbors = list(motif_child_nodes_kl_dist_max(
            current, 
            active_methylation_sequences.pssm(), 
            contig_pssm
        ))

        # Add neighbors to graph
        for next in neighbors:
            if next in motif_graph.nodes:
                # Add only edge if motif -> next
                next_model = motif_graph.nodes[next]["model"]
                score = motif_graph.nodes[next]["score"]
                motif_graph.add_edge(current, next)
            else:
                next_model = motif_model_contig(pileup, contig.sequence, next)

                # Add neighbor to graph
                motif_graph.add_node(next, model=next_model, motif=next, visited=False)
                motif_graph.add_edge(current, next)

                score = scoring_function(next_model, current_model)
                motif_graph.nodes[next]["score"] = score

            # Add neighbor to priority que if has not been visited in this search
            if next not in visisted_nodes:
                priority =  priority_function(next_model, current_model)
                hq.heappush(priority_que, (priority, next))
            
            if score > best_score:
                best_score = score
                best_guess = next
                rounds_since_new_best = 0

        # Stopping criteria
        if rounds_since_new_best >= max_rounds_since_new_best:
            break
    return motif_graph, best_guess


def count_periods_at_start(s):
    count = 0
    for char in s:
        if char == '.':
            count += 1
        else:
            break
    return count
def count_periods_at_end(s):
    s = s[::-1]
    count = 0
    for char in s:
        if char == '.':
            count += 1
        else:
            break
    return count
def nxgraph_to_dataframe(graph):
    return pl.DataFrame({
        "sequence":[i for i in graph.nodes],
        "model":[graph.nodes[i]["model"] for i in graph.nodes],
        "score":[graph.nodes[i]["score"] for i in graph.nodes]
    }).sort("score", descending=True)



if __name__ == "__main__":
    from nanomotif.dataload import load_assembly, load_pileup
    assembly = load_assembly("data/ecoli/assembly.polished.fasta")
    ecoli = load_pileup("data/ecoli/modkit.pileup.bed")
    result = process_sample(assembly, ecoli.pileup, min_read_methylation_fraction = 0.80)



import nanomotif as nm
import polars as pl
from polars import col
import logging as log

def remove_noisy_motifs(motif_df):
    motif_df_clean = []
    for contig, df in motif_df.groupby("contig"):
        motif_strings = df.get_column("motif").to_list()
        positions = df.get_column("mod_position").to_list()
        motifs = [nm.candidate.Motif(motif_string, pos) for motif_string, pos in zip(motif_strings, positions)]
        clean_motifs = []
        for motif in motifs:
            if not motif.have_isolated_bases():
                clean_motifs.append(motif)
        df_clean = df.filter(col("motif").is_in(clean_motifs))
        motif_df_clean.append(df_clean)
    motif_df_clean = pl.concat(motif_df_clean)
    return motif_df_clean

def remove_child_motifs(motifs):
    parent_motifs = []
    for i, motif in enumerate(motifs):
        parent = True
        for j, other in enumerate(motifs):
            if i == j:
                continue
            if motif.sub_string_of(other):
                parent = False
                break
        if parent:
            parent_motifs.append(motif)
    return parent_motifs

def remove_sub_motifs(motif_df):
    motif_df_clean = []
    for contig, df in motif_df.groupby("contig"):
        motif_strings = df.get_column("motif").to_list()
        positions = df.get_column("mod_position").to_list()
        motifs = [nm.candidate.Motif(motif_string, pos) for motif_string, pos in zip(motif_strings, positions)]
        parent_motifs = remove_child_motifs(motifs)
        df_clean = df.filter(col("motif").is_in(parent_motifs))
        motif_df_clean.append(df_clean)
    motif_df_clean = pl.concat(motif_df_clean)
    return motif_df_clean

def merge_motifs_in_df(motif_df, pileup, assembly):
    new_df = []
    for (contig, mod_type), df in motif_df.groupby("contig", "mod_type"):
        # Get list of motifs
        motif_seq = df["motif"].to_list()
        motif_pos = df["mod_position"].to_list()
        motifs = [nm.candidate.Motif(seq, pos) for seq, pos in zip(motif_seq, motif_pos)]

        # Merge motifs
        merged_motifs, pre_merge_motifs = nm.candidate.merge_motifs(motifs)
        
        # Create a new dataframe with the non merged motifs
        if  len(pre_merge_motifs) == 0:
            # No motifs were merged
            new_df.append(df)
            continue
        single_df = df.filter(col("motif").is_in(pre_merge_motifs).not_())
        new_df.append(single_df)
        merged_df = []
        for motif in merged_motifs:
            merged_model = nm.evaluate.motif_model_contig(
                pileup.filter((col("contig") == contig) & (col("mod_type") == mod_type)), 
                assembly.assembly[contig].sequence,
                motif
            )
            merged_df.append(pl.DataFrame({
                "sequence": motif.string,
                "score": -1.,
                "contig": contig,
                "mod_type": mod_type,
                "model": merged_model,
                "motif": motif.string,
                "mod_position": motif.mod_position
            }))
        new_df.append(pl.concat(merged_df))
    new_df = pl.concat(new_df)
    return new_df



import logging as log

def configure_logger(log_file, verbose=False):
    if verbose:
        level=log.DEBUG
    else:
        level=log.INFO
    
    log.basicConfig(
        filename=log_file,
        level=level,
        format="%(asctime)s - %(levelname)s - %(message)s",
        datefmt="%d-%b-%y %H:%M:%S",
    )




import numpy as np
import re
import itertools
import networkx as nx
import logging as log
from scipy.stats import entropy
from nanomotif.constants import *
from nanomotif.seq import regex_to_iupac, iupac_to_regex, reverse_compliment
class Motif(str):
    def __new__(cls, motif_string, *args, **kwargs):
        return str.__new__(cls, motif_string)
    
    def __init__(self, _, mod_position):
        self.mod_position = mod_position
        self.string = self.__str__()

    def from_iupac(iupac_motif):
        """
        Generate Motif from IUPAC seqeunce
        """
    def iupac(self):
        """
        Get IUPAC sequence of motif
        """
        return regex_to_iupac(self.string)

    def sub_motif_of(self, other_motif):
        """
        Check if a motif is in another motif.
        """
        assert isinstance(other_motif, Motif)
        # Strip for redundant posititons (flanking .)
        self_stripped = self.new_stripped_motif()
        other_stripped = other_motif.new_stripped_motif()

        if self_stripped.length() < other_stripped.length():
            return False

        # Split into list of bases
        self_split = self_stripped.split()
        other_split = other_stripped.split()

        index_offset = other_stripped.mod_position - self_stripped.mod_position
        if index_offset > 0:
            return False
        for i, base in enumerate(self_split):
            if i + index_offset < 0:
                continue
            try:
                if set(base).intersection(set(other_split[i + index_offset])) != set(base) and other_split[i + index_offset] != ".": 
                    # Using set to make sure nucleotide order doesn't matter
                    return False
            except IndexError:
                return True
        return True

    def sub_string_of(self, other_motif):
        """
        Check if a motif is in another motif.
        """
        assert isinstance(other_motif, Motif)
        # Strip for redundant posititons (flanking .)
        self_stripped = self.new_stripped_motif()
        other_stripped = other_motif.new_stripped_motif()

        if self_stripped.length() < other_stripped.length():
            return False
        size_difference = self_stripped.length() - other_stripped.length()

        # Split into list of bases
        self_split = self_stripped.split()
        other_split = other_stripped.split()
        
        for i in range (size_difference+1):
            match = True
            for j, base in enumerate(other_split):
                if not set(self_split[j+i]).issubset(set(base)) and base != ".": 
                    # Using set to make sure nucleotide order doesn't matter
                    match = False
            if match:
                return True
        return False 
    
    def distance(self, other_motif):
        assert isinstance(other_motif, Motif)
        self_split = self.split()
        other_motif_split = other_motif.split()

        index_offset = self.mod_position - other_motif.mod_position

        start = (-self.mod_position, -other_motif.mod_position)
        end = (len(self_split) - self.mod_position, len(other_motif_split) - other_motif.mod_position)
        search_size = range(min(start), max(end))
    
        
        distance = 0
        for i in search_size:
            if i < start[0]:
                if other_motif_split[i - start[1]] != ".":
                    distance += 1
            elif i < start[1]:
                if self_split[i - start[0]] != ".":
                    distance += 1
            elif i >= end[0]:
                if other_motif_split[i - start[1]] != ".":
                    distance += 1
            elif i >= end[1]:
                if self_split[i - start[0]] != ".":
                    distance += 1
            else:
                if set(self_split[i - start[0]]) == set(other_motif_split[i - start[1]]):
                    continue
                else:
                    distance += 1
        return distance

    def have_isolated_bases(self, isolation_size=2):
        """
        Check if a motif has isolated bases.
        """
        motif_split = self.split()
        isolated = False
        for pos in range(len(motif_split)):
            if motif_split[pos] == ".":
                continue
            index_start = max(pos - isolation_size, 0)
            index_end = min(pos + isolation_size, len(motif_split) - 1)
            # If all surrounding positions are ".", it is isolated
            if set(motif_split[index_start:pos] + motif_split[pos+1:index_end]) == set(["."]):
                isolated = True
        return isolated
        

    def length(self) -> int:
        """
        Return the length of the motif.
        """
        return len(self.split())
    def trimmed_length(self) -> int:
        """
        Return the length of the motif without dots.
        """
        return len(self.split()) - self.count(".")
    def strip(self, character = ".") -> str:
        """
        Return the motif string with single characters surrounded by dots removed.
        """
        new_motif = self
        return self.lstrip(character).rstrip(character)
    
    def new_stripped_motif(self, character = "."):
        """
        Return the motif string with single characters surrounded by dots removed.
        """
        if re.search("[^\.]", self.string) is None:
            return self
        
        new_position = self.mod_position - re.search("[^\.]", self.string).start()
        new_motif = self.strip(character)

        return Motif(new_motif, new_position)

    def split(self) -> list:
        """
        Split a string into a list of characters, keeping bracketed characters together.

        Returns:
        - list: A list of bases, where bracketed characters are kept together.
        """
        result = []
        i = 0
        while i < len(self):
            if self[i] == "[":
                j = self.find("]", i)
                if j == -1:
                    raise ValueError("Unmatched bracket")
                result.append(self[i:j+1])
                i = j + 1
            else:
                result.append(self[i])
                i += 1
        return result
    
    def one_hot(self) -> np.ndarray:
        """
        Convert the motif to a numpy array.
        """
        split_motif = self.split()
        squence_array = np.zeros((len(split_motif), 4), dtype=int)
        for position, bases in enumerate(split_motif):

            for base in bases:
                if base in BASE_TO_ONE_HOT.keys():
                    squence_array[position, :] += BASE_TO_ONE_HOT[base]
        return squence_array
    
    def reverse_compliment(self):
        """
        Reverse compliment a motif.
        """
        reversed_motif = "".join([COMPLEMENT[base] for base in reversed(self.string)])
        new_position = self.length() - self.mod_position - 1
        return Motif(reversed_motif, new_position)
    
    def merge(self, other):
        assert isinstance(other, Motif)
        mod_pos = self.mod_position
        mod_pos_other = other.mod_position

        # Determine the offset required to align the modification positions
        offset =  mod_pos - mod_pos_other

        # Split the motifs into lists of bases
        motif_split = self.split()
        other_split = other.split()

        motif_right = len(motif_split) - mod_pos 
        other_right = len(other_split) - mod_pos_other
        # Pad the shorter motif with '.' to ensure equal lengths for merging
        if mod_pos < mod_pos_other:
            motif_split = ['.'] * abs(offset) + motif_split
        else:
            other_split = ['.'] * abs(offset) + other_split

        if motif_right < other_right:
            motif_split = motif_split + ['.'] * (other_right - motif_right)
        else:
            other_split = other_split + ['.'] * (motif_right - other_right)
        # Merge the motifs base by base, keeping the modification position aligned
        merged_motif_string = ''.join([self.merge_bases(base, base_other) for base, base_other in zip(motif_split, other_split)])

        # The modification position in the merged motif is the same as in the original motifs
        merged_mod_position = mod_pos + abs(offset)

        # Create and return a new Motif object with the merged motif string and modification position
        return Motif(merged_motif_string, merged_mod_position)
        
    def merge_bases(self, base1, base2):
        canonical_bases = {'A', 'C', 'G', 'T'}
        # Extract bases from brackets if present
        base1_set = set(base1[1:-1]) if '[' in base1 else set(base1)
        base2_set = set(base2[1:-1]) if '[' in base2 else set(base2)

        # Merge the sets of bases
        merged_set = base1_set.union(base2_set)

        # If the merged set has more than one base, enclose them in brackets
        if merged_set == canonical_bases:
            return '.'
        elif "." in merged_set:
            return '.'
        elif len(merged_set.intersection(canonical_bases)) > 1:
            return f"[{''.join(sorted(merged_set.intersection(canonical_bases)))}]"
        else:
            return ''.join(merged_set)  # return as a string without brackets if only one base
    
    def explode_motif(self):
        """
        Explode a motif into all possible combinations of bases.
        """
        exploded_strings = explode_sequence(self.string)
        return [Motif(motif, self.mod_position) for motif in exploded_strings]


def explode_sequence(seq):
    # Splitting the sequence into parts
    parts = []
    temp = ''
    in_brackets = False

    for char in seq:
        if char == '[':
            if temp:
                parts.append([temp])
                temp = ''
            in_brackets = True
            temp += char
        elif char == ']':
            temp += char
            parts.append([temp])
            temp = ''
            in_brackets = False
        else:
            if in_brackets:
                temp += char
            else:
                parts.append([char])

    # Handling the case where the sequence ends with a non-bracketed part
    if temp:
        parts.append([temp])

    # Processing bracketed parts to extract options
    for i, part in enumerate(parts):
        if part[0].startswith('['):
            # Extracting characters between brackets and splitting them into separate options
            parts[i] = list(part[0][1:-1])

    # Generating all combinations
    exploded_sequences = [''.join(combination) for combination in itertools.product(*parts)]

    return exploded_sequences


def merge_motifs(motifs, connectivity_dist=2, min_length=4):
    """
    Merges a Motifs that are closely related. 
    
    Args:
    - motifs (list): List of Motifs to merge.
    - connectivity_dist (int): Maximum edit distance between motifs to be merged.
    - min_length (int): Minimum length of motifs to be merged. 4 Based on estimate from REBASE motifs

    Returns:
    - list: List of merged motifs.
    """
    # Create network where each node is a motif and the edge is the edit distance between them
    distance_graph = nx.Graph()

    # Subset to motif greater than minimum length
    motifs = [motif for motif in motifs if motif.trimmed_length() > min_length]

    # Add nodes and edges to graph
    for i, motif in enumerate(motifs):
        distance_graph.add_node(i, motif=motif)
        for j, other in enumerate(motifs):
            if (i != j):
                distance = motif.distance(other)
                if distance <= connectivity_dist:
                    distance_graph.add_edge(i, j, dist=distance)

    # Find connected components (clusters)
    clusters = list(nx.connected_components(distance_graph))

    # Check for mutual reachability between clusters, all nodes should have an edge to all other nodes in the cluster
    def check_all_nodes_connected(graph):
        n = len(graph.nodes()) 
        expected_edges = n * (n - 1) / 2
        is_complete = len(graph.edges()) == expected_edges
        return is_complete
    clusters = [cluster for cluster in clusters if check_all_nodes_connected(distance_graph.subgraph(cluster))]

    # Now 'clusters' is a list of sets, where each set is a cluster of nodes
    pre_merge_motifs = []
    merged_motifs = []
    for i, cluster in enumerate(clusters, start=1):
        if len(cluster) == 1:
            continue
        log.debug(f'Cluster {i}: {cluster}')
        cluster_motifs = []
        merged_motif = None
        for node in cluster:
            log.debug(f' - Node {node}: {motifs[node]}')
            cluster_motifs.append(motifs[node])
            # Merge motifs
            if merged_motif is None:
                merged_motif = motifs[node]
            else:
                merged_motif = merged_motif.merge(distance_graph.nodes[node]["motif"])

        # check if merged motif gives rise to non observed motifs
        exploded_motifs = merged_motif.explode_motif()
        all_represeneted = True
        for motif1 in exploded_motifs:
            represented = False
            for motif2 in cluster_motifs:
                # Should be a child or equal to atleast one premerge motif
                if motif1.sub_motif_of(motif2):
                    represented = True
                    break
            if not represented:
                all_represeneted = False
                break

        if not all_represeneted:
            log.debug(f' - Merged motif: {merged_motif} does not represent all observed motifs: keeping premerge motifs')
            continue
        log.debug(f' - Merged motif: {merged_motif}')
        pre_merge_motifs += cluster_motifs
        merged_motifs.append(merged_motif)
    return merged_motifs, pre_merge_motifs


def remove_child_motifs(motifs):
    """
    Remove motifs that are children of other motifs in the list.
    """
    new_motifs = []
    for i, motif in enumerate(motifs):
        is_child = False
        for j, other_motif in enumerate(motifs):
            if motif.child_of(other_motif) and i != j:
                is_child = True
        if not is_child:
            new_motifs.append(motif)
    return new_motifs

class MotifTree(nx.DiGraph):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


if __name__ == "__main__":
    motif1 = Motif("AT.G", 0)
    motif2 = Motif("ATCG", 0)
    motif1.sub_string_of(motif2)        


import nanomotif as nm
from nanomotif.logger import configure_logger
import logging as log
import os
from pathlib import Path
import json
os.environ["POLARS_MAX_THREADS"] = "1"
import polars as pl
import warnings

def shared_setup(args, working_dir):
    warnings.filterwarnings("ignore")

    # Set up logging
    LOG_DIR = working_dir + "/logs"
    Path(LOG_DIR).mkdir(parents=True, exist_ok=True)
    configure_logger(LOG_DIR + f"/{args.command}.main.log", args.verbose)

    # Log arguments
    log.debug("Arguments:")
    for arg in vars(args):
        log.debug(f"{arg}: {getattr(args, arg)}")
    
    # Save settings
    with open(working_dir + f"/args.{args.command}.json", "w") as f:
        json.dump(vars(args), f, indent=2)
    
    # Check if output directory exists
    if not os.path.exists(args.output):
        os.makedirs(args.output)
    else:
        log.error(f"Output directory {args.output} already exists")
        return





def find_motifs(args, pileup = None, assembly = None):
    # Run nanomotif
    log.info("Starting nanomotif motif finder")
    if pileup is None:
        log.info("Loading pileup")
        pileup = nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)
    assm_lengths = pl.DataFrame({
        "contig": list(assembly.assembly.keys()),
        "length": [len(contig) for contig in assembly.assembly.values()]
    })
    log.info("Filtering pileup")
    # Filter pileup to contigs with mods, minimum 1 mod per 10kb
    contigs_with_mods = pileup.pileup \
        .groupby(["contig", "mod_type"]) \
        .agg(pl.count()) \
        .join(assm_lengths, on = "contig") \
        .filter(pl.col("count") > pl.col("length")/10000) \
        .get_column("contig").unique().to_list()

    total_contigs = len(assembly.assembly.keys())

    contigs_to_process = [contig for contig in assembly.assembly.keys() if contig in contigs_with_mods]
    pileup = pileup.pileup.filter(pl.col("contig").is_in(contigs_to_process))
    remaining_contigs = pileup.get_column("contig").unique().to_list()
    log.info(f"Processing {len(remaining_contigs)} of {total_contigs} contigs")
    log.info("Identifying motifs")
    
    motifs = nm.evaluate.process_sample_parallel(
            assembly, pileup, 
            threads = args.threads,
            search_frame_size = args.search_frame_size,
            threshold_methylation_confident = args.threshold_methylation_confident,
            threshold_methylation_general = args.threshold_methylation_general,
            threshold_valid_coverage = args.threshold_valid_coverage,
            minimum_kl_divergence = args.minimum_kl_divergence,
            verbose = args.verbose,
            log_dir = args.output + "/logs"
        )

    if motifs is None:
        log.info("No motifs found")
        return

    log.info("Writing motifs")
    def save_motif_df(df, name):
        df.with_columns(
            pl.col("model").apply(lambda x: x._alpha).alias("alpha"),
            pl.col("model").apply(lambda x: x._beta).alias("beta")
        ).drop("model").with_columns([
            pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
        ]) \
        .with_columns([
            pl.col("motif").apply(lambda x: nm.utils.motif_type(x)).alias("motif_type")
        ]) \
        .unique(["motif_iupac", "contig", "mod_type", "mod_position"]) \
        .drop("sequence", "score") \
        .write_csv(args.output + "/" + name + ".tsv", separator="\t")
    os.makedirs(args.output + "/precleanup-motifs/", exist_ok=True)
    save_motif_df(motifs, "precleanup-motifs/motifs-raw")

    log.info("Postprocessing motifs")

    log.info(" - Writing motifs")
    motifs = motifs.filter(pl.col("score") > 0.1)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    save_motif_df(motifs, "precleanup-motifs/motifs-score-filtered")

    log.info(" - Removing sub motifs")
    motifs = nm.postprocess.remove_sub_motifs(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    save_motif_df(motifs, "precleanup-motifs/motifs-score-sub-filtered")

    log.info(" - Removing noisy motifs")
    motifs = nm.postprocess.remove_noisy_motifs(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    save_motif_df(motifs, "precleanup-motifs/motifs-score-sub-noise-filtered")

    log.info(" - Merging motifs")
    motifs = nm.postprocess.merge_motifs_in_df(motifs, pileup, assembly)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    save_motif_df(motifs, "motifs")

    log.info("Done finding motifs")
    return motifs.drop("model")

def score_motifs(args, pileup = None, assembly = None, motifs = None):
    log.info("Starting nanomotif motif scorer")
    if pileup is None:
        log.info("Loading pileup")
        pileup =  nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)
    if motifs is None:
        log.info("Loading motifs")
        motifs = pl.read_csv(args.motifs, separator="\t")
    log.info("Scoring motifs")
    scored_all = nm.scoremotifs.score_sample_parallel(
        assembly, pileup.pileup, motifs,
        threads = args.threads,
        threshold_methylation_general = args.threshold_methylation_general,
        threshold_valid_coverage = 1,
        verbose = args.verbose,
        log_dir = args.output + "/logs"
        )
    scored_all = scored_all \
        .with_columns(
            pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
        ) \
        .with_columns([
            pl.col("motif").apply(lambda x: nm.utils.motif_type(x)).alias("motif_type")
        ]).unique(["motif_iupac", "contig", "mod_type", "mod_position"])
    scored_all.drop("model").write_csv(args.output + "/motifs-scored.tsv", separator="\t")
    return scored_all.drop("model")

def bin_consensus(args, pileup = None, assembly = None, motifs = None, motifs_scored = None):
    bins = pl.read_csv(args.bins, separator="\t", has_header=False) \
        .rename({"column_1":"contig", "column_2":"bin"})
    log.info("Loading motifs-scored")
    if motifs is None:
        motifs = pl.read_csv(args.motifs, separator="\t")
    log.info("Loading motifs")
    if motifs_scored is None:
        motifs_scored = pl.read_csv(args.motifs_scored, separator="\t")
    log.info("Loading pileup")
    if pileup is None:
        pileup =  nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    log.info("Loading assembly")
    if assembly is None:
        assembly = nm.load_assembly(args.assembly)
    motifs_scored = motifs_scored.with_columns([
        pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
    ])
    motifs = motifs.with_columns([
        pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
    ])

    output = nm.bin_consensus.within_bin_motifs_consensus(pileup.pileup, assembly, motifs, motifs_scored, bins)
    output.write_csv(args.output + "/bin-motifs.tsv", separator="\t")

def metagenomic_workflow(args):
    # Check if output directory exists
    pileup = nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    assembly = nm.load_assembly(args.assembly)

    # Find motifs
    motifs = find_motifs(args, pileup=pileup, assembly=assembly)

    # Score all motifs
    scored_all = score_motifs(args, pileup=pileup, assembly=assembly, motifs=motifs)

    # Bin consensus
    bin_consensus(args, pileup=pileup, assembly=assembly, motifs=motifs, motifs_scored=scored_all)

    log.info("Completely done")


def main():
    # Parse arguments
    parser = nm.argparser.create_parser()
    args = parser.parse_args()
    shared_setup(args, args.output)
    if args.command == "find-motifs":
        find_motifs(args)
    elif args.command == "score-motifs":
        score_motifs(args)
    elif args.command == "bin-consensus":
        bin_consensus(args)
    elif args.command == "complete-workflow":
        metagenomic_workflow(args)
    #elif args.command == "clean-bins":
    #    pass
    #elif args.command == "associate-mges":
    #    pass
    else:
        parser.print_help()
        exit()
if __name__ == "__main__":
    main()



from nanomotif.dataload import *
import nanomotif.evaluate, nanomotif.model, nanomotif.utils, nanomotif.seq, nanomotif.main, nanomotif.argparser, nanomotif.constants, nanomotif.candidate, nanomotif.postprocess, nanomotif.scoremotifs, nanomotif.bin_consensus



from scipy.stats import entropy
import numpy as np
import warnings
import re
import random
import nanomotif.utils as utils
from nanomotif.constants import *
import matplotlib.pyplot as plt
from matplotlib import cm

class Assembly():
    def __init__(self, assembly: dict):
        self.assembly = {name:DNAsequence(seq) for name, seq in assembly.items()}
    
    def __getitem__(self, key):
        return self.assembly[key]
    
    def __repr__(self):
        return f"Assembly with {len(self.assembly)} contigs"

class DNAsequence:
    bases = ["A", "T", "G", "C"]
    iupac = ["A", "T", "G", "C", "R", "Y", "S", "W", "K", "M", "B", "D", "H", "V", "N"]
    iupac_dict = {
        "A": ["A"], "T": ["T"], "C": ["C"], "G": ["G"],
        "R": ["A", "G"], "Y": ["C", "T"], 
        "S": ["G", "C"], "W": ["A", "T"], 
        "K": ["G", "T"], "M": ["A", "C"],
        "B": ["C", "G", "T"],
        "D": ["A", "G", "T"],
        "H": ["A", "C", "T"],
        "V": ["A", "C", "G"],
        "N": ["A", "T", "C", "G"]
    }
    complement = {
        "A": "T", "T": "A", "G": "C", "C": "G", 
        "N": "N", "R": "Y", "Y": "R", "S": "S", 
        "W": "W", "K": "M", "M": "K", "B": "V", 
        "D": "H", "H": "D", "V": "B"
    }
    base_to_vector = {
        "A": [1, 0, 0, 0],
        "T": [0, 1, 0, 0],
        "G": [0, 0, 1, 0],
        "C": [0, 0, 0, 1],
        "N": [1, 1, 1, 1],
        ".": [1, 1, 1, 1]
    }

    base_int = {"A":1, "T":2, "G":3, "C":4, "N":5, "R":6, "Y":7, "S":8, "W":9, "K":10, "M":11, "B":12, "D":13, "H":14, "V":15}
    int_base = {i: n for n, i in base_int.items()}

    def __init__(self, sequence):
        self._check_sequence(sequence)
        sequence = sequence.upper()
        self.sequence = sequence
    
    @property
    def sequence(self):
        return self._sequence

    @sequence.setter
    def sequence(self, value):
        self.sequence_alphabet = list(set(value))
        self._check_sequence(value)
        self._sequence = value

    def _check_sequence(self, sequence):
        assert isinstance(sequence, str), "DNA sequence must be a str"
        assert len(sequence) > 0, "DNA sequence must not be empty"
        assert all(letter in self.iupac for letter in list(set(sequence))), f"DNA sequence must be a nucleotide sequence of {''.join(self.iupac)}"

    
    def __getitem__(self, key):
        return self.sequence[key]

    def __len__(self):
        return len(self.sequence)

    def __iter__(self):
        return iter(self.sequence)
    
    def __repr__(self):
        return f"DNAsequence() | Number of sequence: {len(self)} | Unique Bases: {self.sequence_alphabet}"

    def __add__(self, other):
        if isinstance(other, DNAsequence):
            return DNAsequence(self.sequence + other.sequence)
        elif isinstance(other, str):
            return DNAsequence(self.sequence + other)
        else:
            return NotImplemented

    def __mul__(self, other):
        try:
            return DNAsequence(self.sequence * other)
        except:
            return NotImplemented

    def __eq__(self, other):
        if isinstance(other, DNAsequence):
            return self.sequence == other.sequence
        elif isinstance(other, str):
            return self.sequence == other
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, DNAsequence):
            return self.sequence != other.sequence
        elif isinstance(other, str):
            return self.sequence != other
        else:
            return NotImplemented

    def __lt__(self, other):
        if isinstance(other, int):
            return len(self) < other
        elif isinstance(other, (DNAsequence, str)):
            return len(self) < len(other)
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, int):
            return len(self) <= other
        elif isinstance(other, (DNAsequence, str)):
            return len(self) <= len(other)
        else:
            return NotImplemented

    def __gt__(self, other):
        if isinstance(other, int):
            return len(self) > other
        elif isinstance(other, (DNAsequence, str)):
            return len(self) > len(other)
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, int):
            return len(self) >= other
        elif isinstance(other, (DNAsequence, str)):
            return len(self) >= len(other)
        else:
            return NotImplemented
    
    def sample_at_index(self, index: int, padding: int):
        """
        Randomly sample sequence around index in DNAsequence.

        Parameters
        ----------
        index : int
            Index to sample at.
        padding : int
            Number of position to include on each side

        Returns
        -------
        sequences : DNAsequence
            Sequence of length 2 * padding + 1
        """
        if index < padding:
            raise ValueError(f"Index {index} is too small for padding {padding}")
        if index > len(self) - padding:
            raise ValueError(f"Index {index} is too large for padding {padding}")
        return DNAsequence(self.sequence[(index - padding):(index + padding + 1)])

    def sample_at_indices(self, indices: list, padding: int):
        """
        Randomly sample sequence around indices in DNAsequence.

        Parameters
        ----------
        indices : list
            Indices to sample at.
        padding : int
            Number of position to include on each side

        Returns
        -------
        sequences : EqualLengthDNASet
            Sequences of length 2 * padding + 1
        """
        indices_checked = [index for index in indices if (index > padding) and (index < (len(self) - padding))]
        return EqualLengthDNASet([self.sample_at_index(index, padding) for index in indices_checked])

    def sample_subsequence(self, length):
        """Randomly sample a subsequence of a given length"""
        if length > len(self):
            raise ValueError(f"Cannot sample a subsequence of length {length} from a sequence of length {len(self)}")
        start_index = random.randint(0, len(self) - length)
        return DNAsequence(self.sequence[start_index:start_index + length])

    def sample_n_subsequences(self, length: int, n: int):
        """Randomly sample n sequences of a given length"""
        return EqualLengthDNASet([self.sample_subsequence(length=length) for _ in range(n)])

    def find_subsequence(self, subsequence: str):
        """
        Find all occourances of subseqeunce in DNAsequence
        """
        p = len(subsequence)
        t = len(self.sequence)
        subsequence_hash = hash(subsequence)
        sequence_hash = hash(self.sequence[:p])
        indices = []
        for i in range(t - p + 1):
            if subsequence_hash == sequence_hash:
                if subsequence == self.sequence[i:i+p]:
                    indices.append(i)
            if i < t - p:
                sequence_hash = hash(self.sequence[i+1:i+p+1])
        
        return np.array(indices)

    def sample_at_subsequence(self, padding: int, subsequence: str):
        """
        Randomly sample sequence around subsequence in DNAsequence.

        Parameters
        ----------
        padding : int
            Number of position to include on each side
        subsequence : str
            Subsequence to sample at.

        Returns
        -------
        sequences : DNAsequence
            Sequence of length 2 * padding + len(subsequence)
        """
        subsequence_index = self.find_subsequence(subsequence)
        subsequence_index = subsequence_index[subsequence_index >= padding]
        subsequence_index = subsequence_index[subsequence_index < len(self.sequence) - padding]
        if subsequence_index.shape[0] == 0:
            warnings.warn("Not occourances found " + subsequence + " found")
        else:
            sample_index = np.random.choice(subsequence_index, 1)[0]
            return DNAsequence(self.sequence[(sample_index - padding):(sample_index + padding + len(subsequence))])

    def sample_n_at_subsequence(self, padding: int, subsequence: str, n: int):
        """
        Randomly sample sequence around subsequence in DNAsequence.

        Parameters
        ----------
        padding : int
            Number of position to include on each side
        subsequence : str
            Subsequence to sample at.
        n: int
            number of sequences to sample (all unique)

        Returns
        -------
        sequences : EqualLengthDNASet
            Sequences of length 2 * padding + len(subsequence)
        """
        subsequence_index = self.find_subsequence(subsequence)
        subsequence_index = subsequence_index[subsequence_index >= padding]
        subsequence_index = subsequence_index[subsequence_index < (len(self.sequence) - padding)]
        if subsequence_index.shape[0] < n:
            warnings.warn("Only found " + str(subsequence_index.shape[0]) + " occourances, sampling " + str(subsequence_index.shape[0]) + " sequences")
            n = subsequence_index.shape[0]
        sample_index = np.random.choice(subsequence_index, n, replace=False)
        sampled_sequences = [DNAsequence(self.sequence[(i - padding):(i + padding + len(subsequence))]) for i in sample_index]
        return EqualLengthDNASet(sampled_sequences)

    def reverse_complement(self):
        """Return the reverse complement of the sequence."""
        return DNAsequence("".join([self.complement[base] for base in reversed(self.sequence)]))

    def count(self, subsequence):
        """Count the number of occurrences of a subsequence in the sequence."""
        return self.sequence.count(subsequence)
    
    def gc(self) -> list:
        '''
        Returns the GC content of a sequence

        Returns
        -------
        Float
            GC content of sequence

        Examples
        --------
        >>> DNAsequence("ATCG").gc()
        0.5
        '''
        return (self.count("G") + self.count("C")) / len(self)





class EqualLengthDNASet:
    def __init__(self, sequences: list):
        self._check_sequences(sequences)
        self.sequences = sequences
        self.sequence_length = len(sequences[0])

    @property
    def sequences(self):
        return self._sequences

    @sequences.setter
    def sequences(self, value):
        self._check_sequences(value)
        self._sequences = value
    
    def __add__(self, other):
        if isinstance(other, EqualLengthDNASet):
            return EqualLengthDNASet(self.sequences + other.sequences)
        elif isinstance(other, list):
            return EqualLengthDNASet(self.sequences + other)
        else:
            return NotImplemented

    def _check_sequences(self, sequences):
        assert isinstance(sequences, list), "DNA sequences must be a list"
        assert len(sequences) > 0, "DNA sequences must not be empty"
        assert all(isinstance(seq, DNAsequence) for seq in sequences), "DNA sequences must be a list of strings"
        assert all(len(seq) > 0 for seq in sequences), "DNA sequences must not contain empty strings"
        assert utils.all_lengths_equal(sequences), "All sequences must be of equal length"

    def __getitem__(self, key):
        return EqualLengthDNASet(self.sequences[key])

    def levenshtein_distances(self) -> np.ndarray:
        """
        All vs. all levenshtein distances between sequences.

        Returns
        -------
        np.ndarray
            A matrix of edit distances between all sequences

        Examples
        --------
        >>> EqualLengthDNASet(["ATCG", "GCTA", "TACT", "AGCT"]).levenshtein_distances()
        array([[0, 4, 3, 2],
               [4, 0, 3, 2],
               [3, 3, 0, 2],
               [2, 2, 2, 0]], dtype=int16)
        """
        lengths = [len(s) for s in self]
        assert utils.all_equal(lengths), "All sequences must be of equal length"

        n = len(self)
        distances = np.empty(shape = (n, n), dtype=np.int16)
        for i in range(0, n):
            for j in range(i, n):
                d = editdistance.eval(self[i], self[j])
                distances[i, j] = d
                distances[j, i] = d
        return distances

    def reverse_compliment(self):
        """Return the reverse complement of the sequence."""
        return EqualLengthDNASet([seq.reverse_complement() for seq in self.sequences])

    def pssm(self, pseudocount=0, only_canonical: bool = True):
        """
        Calculate the positional frequency of each nucleotide for all sequences.

        Parameters
        ----------
        pseudocount : float, optional
            Pseudocount to be added to each nucleotide count. Default is 0.5.

        Returns
        -------
        np.ndarray
            A numpy array of positional frequencies.
        """
        if only_canonical:
            bases = self.sequences[0].bases
        else:
            bases = self.sequences[0].iupac

        n = len(self.sequences)

        pssm = []
        for nuc in bases:
            pssm_nuc = []
            for i in range(self.sequence_length):
                count = 0
                for seq in self.sequences:
                    if seq[i] == nuc:
                        count += 1
                pssm_nuc.append((count + pseudocount)/n)
            pssm.append(pssm_nuc)
        return np.array(pssm)

    def kl_divergence(self, pssm):
        """
        Calculate positional Kullback-Liebler divergence from sequences to provided PSSM.
        Parameters
        ----------
        pssm : Position-specific scoring matrix
            PSSM to calculate distances from
        
        Returns
        -------
        np.ndarray
            A numpy array of Kullback-Leibler divergence.
        """
        assert pssm.shape == (4, self.sequence_length), "Shape does not match (should be [4, sequence_length])"
        return entropy(self.pssm(), pssm)

    def get_sequences_with_match(self, pattern: str):
        """
        Remove all sequences that contain a given pattern.

        Parameters
        ----------
        pattern : str
            pattern to remove

        Returns
        -------
        EqualLengthDNASet with sequences containing pattern
        """
        return EqualLengthDNASet([seq for seq in self.sequences if bool(re.match(pattern, seq.sequence))])

    def get_sequences_without_match(self, pattern: str):
        """
        Remove all sequences that contain a given pattern.

        Parameters
        ----------
        pattern : str
            pattern to remove

        Returns
        -------
        EqualLengthDNASet with sequences containing pattern
        """
        sequences = [seq for seq in self.sequences if not bool(re.match(pattern, seq.sequence))]
        if len(sequences) == 0:
            warnings.warn("No sequences left after filtering")
            return None
        return EqualLengthDNASet(sequences)

    def convert_to_DNAarray(self):
        """
        Convert EqualLengthDNASet to DNAarray
        """
        return DNAarray([[DNAsequence.base_to_vector[base] for base in sequence] for sequence in self.sequences])
    
    def plot_pssm(self, ax=None, center_x_axis=True):
        """
        Method to plot the positional conservation of DNA sequences in the object.

        Parameters:
        ax (matplotlib.axes.Axes, optional): An existing Axes object to draw the plot onto, 
                                            default is None in which case a new figure will be created.
        Returns:
        None

        This method generates a plot showing the conservation of sequences in the object.
        It plots the positional frequencies of each nucleotide.
        """
        
        # If no Axes object provided, create a new figure and axes
        if ax is None:
            _, ax = plt.subplots()

        # Plot positional frequencies for each nucleotide
        int_base = self.sequences[0].int_base
        for i in range(1,5):
            ax.plot(self.pssm()[i-1], label=f"{int_base[i]}", linewidth=2, c=cm.cubehelix(i/5))

        # Configure x-axis labels and ticks
        if center_x_axis:
            n_labs = min(4, (self.sequence_length-1)//2) 
            window = (self.sequence_length-1)//2
            x_axis_step = max(window//n_labs, 1)
            labs_positive = np.arange(0, x_axis_step * n_labs + 1, x_axis_step)
            tick_label = np.concatenate((-labs_positive, labs_positive))
            tick_position = tick_label + window
            ax.set_xticks(tick_position)
            ax.set_xticklabels(tick_label)
            ax.set_xlabel("Relative Position")
        else:
            ax.set_xticks(np.arange(0, self.sequence_length, 1))
            ax.set_xlabel("Position")
        # Set title and labels
        ax.set_title(f"Number of sequences: {len(self.sequences)}")
        ax.set_ylabel("Frequency")

        return ax

class DNAarray(np.ndarray):
    """
    A numpy array of DNA sequences
    """
    base_to_vector = {
        "A": [1, 0, 0, 0],
        "T": [0, 1, 0, 0],
        "G": [0, 0, 1, 0],
        "C": [0, 0, 0, 1],
        "N": [1, 1, 1, 1],
        ".": [1, 1, 1, 1]
    }
    def __new__(cls, input_array):
        obj = np.asarray(input_array).view(cls)
        return obj
    
    def __array_finalize__(self, obj):
        if obj is None: return
    
    def filter_sequence_matches(self, sequence, keep_matches=True):
        """
        Filter the array based on sequence matches

        Parameters
        ----------
        sequence : np.ndarray
        keep_matches : bool, optional
            Whether to keep matches or non-matches, by default True

        Returns
        -------
        DNAarray
            The filtered array
        """
        assert isinstance(sequence, np.ndarray), "Sequence must be a numpy array"
        assert sequence.shape == self.shape[1:3], "Sequence must have the same length as sequences in the array"

        if keep_matches:
            result = self[np.all(self <= sequence, axis=(1,2)), :]
        else:
            result = self[np.any(self > sequence, axis=(1,2)), :]
        if result.shape[0] == 0:
            warnings.warn("No sequences left after filtering")
            return None
        return DNAarray(result)
    
    def pssm(self):
        """
        Calculate the position specific scoring matrix for the array. 
        The PSSM is the sum of the array along the first axis (sequences) divided by the number of sequences, 
        e.g. the positional frequency of bases in all of the sequences in the array.

        Returns
        -------
        np.ndarray
            The PSSM
        """
        return self.sum(axis = 0).transpose() / self.shape[0]

def regex_to_iupac(regex):
    # Dictionary to map sorted nucleotide combinations to IUPAC codes
    mapping = {
        "A": "A",
        "T": "T",
        "C": "C",
        "G": "G",
        "AG": "R",
        "CT": "Y",
        "CG": "S",
        "AT": "W",
        "GT": "K",
        "AC": "M",
        "CGT": "B",
        "AGT": "D",
        "ACT": "H",
        "ACG": "V"
    }

    # Function to sort characters inside brackets and replace with IUPAC code
    def replace_with_iupac(match):
        sorted_seq = "".join(sorted(match.group(1)))
        return mapping.get(sorted_seq, "")

    # Replace all [xy] patterns with their corresponding IUPAC code
    iupac_seq = re.sub(r'\[([A|T|C|G]+)\]', replace_with_iupac, regex)

    # Replace remaining dots with N for any nucleotide
    iupac_seq = iupac_seq.replace(".", "N")
    
    return iupac_seq

def iupac_to_regex(iupac):
    # Dictionary to map sorted nucleotide combinations to IUPAC codes
    mapping = {
        "A": "A",
        "T": "T",
        "C": "C",
        "G": "G",
        "R": "AG",
        "Y": "CT",
        "S": "CG",
        "W": "AT",
        "K": "GT",
        "M": "AC",
        "B": "CGT",
        "D": "AGT",
        "H": "ACT",
        "V": "ACG",
        "N": "."
    }

    # Function to sort characters inside brackets and replace with IUPAC code
    def replace_with_iupac(match):
        regex_seq = mapping.get(match.group(0))
        if len(regex_seq) == 1:
            return regex_seq
        else:
            return "[" + mapping.get(match.group(0)) + "]"

    # Replace all [xy] patterns with their corresponding IUPAC code
    regex_seq = re.sub(r'.', replace_with_iupac, iupac)
    return regex_seq



def regex_to_iupac_n(regex):
    # Dictionary to map sorted nucleotide combinations to IUPAC codes
    mapping = {
        "A": "A",
        "T": "T",
        "C": "C",
        "G": "G",
        "AG": "R",
        "CT": "Y",
        "CG": "S",
        "AT": "W",
        "GT": "K",
        "AC": "M",
        "CGT": "B",
        "AGT": "D",
        "ACT": "H",
        "ACG": "V"
    }

    # Function to sort characters inside brackets and replace with IUPAC code
    def replace_with_iupac(match):
        sorted_seq = "".join(sorted(match.group(1)))
        return mapping.get(sorted_seq, "")

    # Replace all [xy] patterns with their corresponding IUPAC code
    iupac_seq = re.sub(r'\[([A|T|C|G]+)\]', replace_with_iupac, regex)

    # Replace remaining dots with N for any nucleotide
    iupac_seq = iupac_seq.replace(".", "N")

    # Replace consecutive N's with N followed by the count (only if 2 or more)
    def replace_ns(match):
        length = len(match.group(0))
        return r'$\mathregular{N_' + str(length) + '}$' if length > 4 else 'N'

    iupac_seq = re.sub(r'N+', replace_ns, iupac_seq)
    
    return iupac_seq


def reverse_compliment(seq):
    """Return the reverse complement of the sequence."""
    return "".join([COMPLEMENT[base] for base in reversed(seq)])


# Data loading functionalities
import polars as pl
from .seq import Assembly
from .feature import Pileup

def load_fasta(path, trim_names=False, trim_character=" ") -> dict:
    """
    Reads a fasta file and returns a dictionary with the contig names as 
    keys and the sequences as values
    """
    with open(path, 'r') as f:
        lines = f.readlines()
    data = {}
    active_sequence_name = "no_header"
    for line in lines:
        line = line.rstrip()
        if line[0] == '>':
            active_sequence_name = line[1:]
            if trim_names:
                active_sequence_name = active_sequence_name.split(trim_character)[0]
            if active_sequence_name not in data:
                data[active_sequence_name] = ''
        else:
            data[active_sequence_name] += line
    return data

def load_pileup(path: str, threads: int = 1, min_fraction: float = 0):
    """
    Load pileup file from path to pileup.bed output of modkit pileup
    """
    pileup = (
        pl.scan_csv(path, separator = "\t", has_header = False)
        .filter(pl.col("column_11") > min_fraction*100)
        .filter(pl.col("column_10") / (pl.col("column_10") + pl.col("column_17")) > 0.3)
        .select(["column_1", "column_2","column_4", "column_6", "column_11", "column_10"])
        .with_columns(pl.col("column_11") / 100)
        .collect()
    )
    pileup = pileup.rename({"column_1":"contig", "column_2": "position", "column_4": "mod_type", "column_6": "strand", "column_11": "fraction_mod", "column_10":"Nvalid_cov"})
    return Pileup(pileup)

def load_assembly(path: str):
    """
    Load assembly from path to fasta file
    """
    return Assembly(load_fasta(path))



import argparse
from nanomotif._version import __version__


def  create_parser():
    parser = argparse.ArgumentParser(description="Motif identification and utilisation commands")
    parser.add_argument("--version", action="version", version="%(prog)s {}".format(__version__))

    subparsers = parser.add_subparsers(help="sub-command help", dest="command")

    ###########################################################################
    # Complete workflow
    parser_complete_workflow = subparsers.add_parser('complete-workflow', help='Run the complete workflow')

    parser_complete_workflow.add_argument("assembly", type=str, help="Assembly file.")
    parser_complete_workflow.add_argument("pileup", type=str, help="Modkit pileup file.")
    parser_complete_workflow.add_argument("bins", type=str, help="File specifying to which bin contigs belong. (tsv file with no header and columns: contig, bin)")
    parser_complete_workflow.add_argument("output", type=str, help="Output directory.")

    parser_complete_workflow.add_argument("-t", "--threads", type=int, default=1, help="Number of threads to use. Default is 1.")
    parser_complete_workflow.add_argument("-v", "--verbose", action="store_true", help="Increase output verbosity. (set logger to debug level)")

    parser_complete_workflow.add_argument("--search_frame_size", type=int, default=40, help="Length of the sequnces sampled around methylatyion sites. Default: %(default)s")
    parser_complete_workflow.add_argument("--threshold_methylation_confident", type=float, default=0.8, help="Minimum fraction of reads that must be methylated to be considered in motif search. Default: %(default)s")
    parser_complete_workflow.add_argument("--threshold_methylation_general", type=float, default=0.5, help="Minimum fraction of reads that must be methylated for a position to be considered modified. Default: %(default)s")
    parser_complete_workflow.add_argument("--threshold_valid_coverage", type=int, default=5, help="Minimum valid base coverage for a position to be considered. Default: %(default)s")
    parser_complete_workflow.add_argument("--minimum_kl_divergence", type=float, default=0.2, help="Minimum KL-divergence for a position to considered for expansion in  motif search. Default: %(default)s")



    ###########################################################################
    # Find Motifs
    parser_find_motifs = subparsers.add_parser('find-motifs', help='Find motifs in contigs of the assembly')

    parser_find_motifs.add_argument("assembly", type=str, help="Path to the assembly file.")
    parser_find_motifs.add_argument("pileup", type=str, help="Path to the modkit pileup file.")
    parser_find_motifs.add_argument("output", type=str, help="Path to the output file.")


    parser_find_motifs.add_argument("-t", "--threads", type=int, default=1, help="Number of threads to use. Default is 1.")
    parser_find_motifs.add_argument("-v", "--verbose", action="store_true", help="Increase output verbosity. (set logger to debug level)")

    parser_find_motifs.add_argument("--search_frame_size", type=int, default=40, 
                                    help="Length of the sequnces sampled around confident methylatyion sites. Default: %(default)s")
    parser_find_motifs.add_argument("--threshold_methylation_confident", type=float, default=0.8, 
                                    help="Minimum fraction of reads that must be methylated at a position for the position to be considered confiently methylated. These position are used to search for candidate motifs. Default: %(default)s")
    parser_find_motifs.add_argument("--threshold_methylation_general", type=float, default=0.6, 
                                    help="Minimum fraction of reads that must be methylated at a position for the position to be methylated. These position are used for counting number of methylated position of a motif. Default: %(default)s")
    parser_find_motifs.add_argument("--threshold_valid_coverage", type=int, default=5, 
                                    help="Minimum valid base coverage for a position to be considered. Default: %(default)s")
    parser_find_motifs.add_argument("--minimum_kl_divergence", type=float, default=0.2, 
                                    help="Minimum KL-divergence for a position to considered for expansion in  motif search. Higher value means less exhaustive, but faster search. Default: %(default)s")

    ###########################################################################
    # Score motifs
    parser_score_motifs = subparsers.add_parser('score-motifs', help='Find degree of methylation of all identified motifs for all contigs in the assembly')

    parser_score_motifs.add_argument("assembly", type=str, help="Path to the assembly file.")
    parser_score_motifs.add_argument("pileup", type=str, help="Path to the modkit pileup file.")
    parser_score_motifs.add_argument("motifs", type=str, help="Path to the motifs file.")
    parser_score_motifs.add_argument("output", type=str, help="Path to the output file.")

    parser_score_motifs.add_argument("-t", "--threads", type=int, default=1, help="Number of threads to use. Default: %(default)s")
    parser_score_motifs.add_argument("-v", "--verbose", action="store_true", help="Increase output verbosity")
    parser_score_motifs.add_argument("--threshold_methylation_general", type=float, default=0.5, help="Minimum fraction of reads that must be methylated for a position to be considered modified. Default: %(default)s")
    ###########################################################################
    # Bin consensus
    parser_bin_consensus = subparsers.add_parser('bin-consensus', help='Find consensus motif in each bin')

    parser_bin_consensus.add_argument("bins", type=str, help="TSV file specifying which bin contigs belong.")
    parser_bin_consensus.add_argument("assembly", type=str, help="Path to the assembly file.")
    parser_bin_consensus.add_argument("pileup", type=str, help="Path to the modkit pileup file.")
    parser_bin_consensus.add_argument("motifs", type=str, help="Path to the motifs file.")
    parser_bin_consensus.add_argument("motifs_scored", metavar="motifs-scored", type=str, help="Path to the motif-scored file.")
    parser_bin_consensus.add_argument("output", type=str, help="Path to the output file.")

    parser_bin_consensus.add_argument("-t", "--threads", type=int, default=1, help="Number of threads to use. Default: %(default)s")
    parser_bin_consensus.add_argument("-v", "--verbose", action="store_true", help="Increase output verbosity")
    parser_bin_consensus.add_argument("--threshold_methylation_general", type=float, default=0.6, help="Minimum fraction of reads that must be methylated for a position to be considered modified. Default: %(default)s")

#    ###########################################################################
#    # Clean bins using motifs
#    parser_clean_bins = subparsers.add_parser('clean-bins', help='Clean bins using methylation motifs [NOT IMPLEMENTED]')
#
#    ###########################################################################
#    # place MGEs
#    parser_associate_mges = subparsers.add_parser('associate-mges', help='Associate plasmids and viruses with bin of similair motif profile [NOT IMPLEMENTED]')

    return parser
    


from setuptools import setup, find_packages
exec(open('nanomotif/_version.py').read())
setup(
    name='nanomotif',
    version=__version__,
    description='Identifying methlyation motifs in nanopore data',
    author='AAU_DarkScience',
    author_email='shei@bio.aau.com',
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        "wheel",
        "requests",
        "numpy==1.24.4",
        "pandas==2.0.2",
        "polars==0.20",
        "seaborn==0.12.2",
        "scipy==1.10.1",
        "networkx==3.1",
        "progressbar2==3.53.1"
    ],
    entry_points={
            'console_scripts': [
                  'nanomotif = nanomotif.main:main'
            ]
    }
)


import polars as pl
import nanomotif as nm


def within_bin_motifs_consensus(pileup, assembly, motifs, motifs_scored, bins):
    motifs = motifs.with_columns([
        pl.lit(True).alias("directly_detected")
    ]).with_columns([
        pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
    ]).with_columns([
        pl.col("motif_iupac").apply(lambda x: nm.seq.iupac_to_regex(x)).alias("motif")
    ])

    motifs_scored = motifs_scored.with_columns([
        pl.lit(True).alias("directly_detected")
    ]).with_columns([
        pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
    ]).with_columns([
        pl.col("motif_iupac").apply(lambda x: nm.seq.iupac_to_regex(x)).alias("motif")
    ]) \
    .join(bins, on="contig", how = "outer") \
    .with_columns(
        pl.when(pl.col("bin").is_null()).then(False).otherwise(pl.lit(True)).alias("binned")
    ).join(
        motifs.drop("sequence",	"score", "alpha", "beta", "motif"), on=["contig", "motif_iupac", "mod_type", "mod_position"], how="left"
    ).with_columns(
        pl.when(pl.col("directly_detected").is_null()).then(pl.lit(False)).otherwise(pl.col("directly_detected")).alias("directly_detected")
    )

    # keep directly detected motifs
    bin_consensus = motifs_scored.groupby("bin", "motif", "mod_position", "mod_type") \
        .apply(lambda group: group.filter(
            pl.col("directly_detected").any()
        )).with_columns(
            pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif_iupac")
        )
    # Merge motifs
    print(bin_consensus)
    print("Merging motifs")
    merged_motifs = merge_motifs_in_df(bin_consensus.select(["contig","mod_type", "alpha", "beta","motif","mod_position",]), pileup, assembly)
    # Remove submotifs
    merged_motifs = nm.postprocess.remove_sub_motifs(merged_motifs)
    
    print(merged_motifs)
    print("Merging motifs2")
    # Join merged motifs with unmerged motifs
    merged_motifs = merged_motifs \
        .join(bin_consensus, on=["contig", "motif", "mod_position", "mod_type"], how="left") \
        .with_columns(
            pl.col("motif").apply(lambda x: nm.seq.regex_to_iupac(x)).alias("motif")
        )

    print(merged_motifs)
    print("Calculating mean")

    bin_consensus = merged_motifs.unique(["bin", "motif", "mod_position", "mod_type", "contig"]) \
        .with_columns(
            (pl.when(pl.col("mean") > 0.5).then(True).otherwise(False)).alias("count_mean_threshold")
        )
    print(bin_consensus)
    print("Calculating consensus")
    bin_consensus = bin_consensus \
        .groupby("bin", "motif", "mod_position", "mod_type") \
        .agg(
            pl.col("alpha").sum().alias("alpha_sum"),
            pl.col("beta").sum().alias("beta_sum"),
            pl.col("directly_detected").sum().alias("count_directly_detected"),
            pl.col("count_mean_threshold").sum().alias("count_mean_threshold"),
            pl.col("contig").count().alias("contig_count")
        ).with_columns( 
            (pl.col("alpha_sum") / (pl.col("alpha_sum") + pl.col("beta_sum"))).alias("mean_sum"),
            (pl.col("count_mean_threshold") / pl.col("contig_count").cast(pl.Float32)).alias("fraction_mean_threshold"),
        )
    bin_consensus = bin_consensus \
        .with_columns(
            pl.col("motif").apply(lambda x: nm.utils.motif_type(x)).alias("motif_type")
        )
    print(bin_consensus)
    print("Filtering")
        

    output = bin_consensus.select(["bin", "motif", "mod_position", "mod_type", "contig_count", "count_mean_threshold", "motif_type", "alpha_sum", "beta_sum", "mean_sum"]) \
        .rename({"alpha_sum":"methylated_count", "beta_sum":"non_methylated_count", "count_mean_threshold":"contigs_with_motif", "mean_sum":"mean_methylation_bin"}) \
        .filter(pl.col("contigs_with_motif") > 0) \
        .filter(pl.col("methylated_count") > 10) \
        .with_columns(
            pl.when(pl.col("bin").is_null()).then("unbinned").otherwise(pl.col("bin")).alias("bin")
        )
    print(output)
    return output



def merge_motifs_in_df(motif_df, pileup, assembly):
    new_df = []
    for (contig, mod_type), df in motif_df.groupby("contig", "mod_type"):
        # Get list of motifs
        motif_seq = df["motif"].to_list()
        motif_pos = df["mod_position"].to_list()
        motifs = [nm.candidate.Motif(seq, pos) for seq, pos in zip(motif_seq, motif_pos)]

        # Merge motifs
        merged_motifs, pre_merge_motifs = nm.candidate.merge_motifs(motifs)
        
        # Create a new dataframe with the non merged motifs
        if  len(pre_merge_motifs) == 0:
            # No motifs were merged
            new_df.append(df)
            continue
        single_df = df.filter(pl.col("motif").is_in(pre_merge_motifs).not_())
        new_df.append(single_df)
        merged_df = []
        for motif in merged_motifs:
            merged_model = nm.evaluate.motif_model_contig(
                pileup.filter((pl.col("contig") == contig) & (pl.col("mod_type") == mod_type)), 
                assembly.assembly[contig].sequence,
                motif
            )
            merged_df.append(pl.DataFrame({
                "contig": contig,
                "mod_type": mod_type,
                "alpha": merged_model._alpha,
                "beta": merged_model._beta,
                "motif": motif.string,
                "mod_position": motif.mod_position
            }))
        new_df.append(pl.concat(merged_df))
    new_df = pl.concat(new_df)

    return new_df




import polars as pl
import multiprocessing
import os
import warnings
from multiprocessing import get_context
import logging as log

from nanomotif import candidate, evaluate
from nanomotif.parallel import update_progress_bar
from nanomotif.logger import configure_logger

def score_contig_with_motifs(contig, mod_type, pileup, sequence, motifs):
    log.info(f"Scoring contig {contig}, modtype {mod_type}")
    motifs = motifs.select("motif", "mod_position", "mod_type").unique()
    motifs = motifs.filter(pl.col("mod_type") == mod_type)
    motifs_str = motifs.get_column("motif").to_list()
    positions = motifs.get_column("mod_position").to_list()
    types = motifs.get_column("mod_type").to_list()

    models = []
    for motif, position, mod_type in zip(motifs_str, positions, types):
        log.debug(f"Scoring motif {motif}, {position}")
        model = evaluate.motif_model_contig(
            pileup,
            sequence, 
            candidate.Motif(motif, position)
        )
        models.append(model)
    log.debug(f"{contig}-{mod_type}, preparing results")
    result = pl.DataFrame({
            "motif": motifs_str,
            "mod_position": positions,
            "mod_type": types,
            "model": models
        })
    log.debug(f"{contig}-{mod_type}, postprocessing results")
    result = result.with_columns([
            pl.col("model").map_elements(lambda x: x._alpha).alias("alpha"),
            pl.col("model").map_elements(lambda x: x._beta).alias("beta"),
            pl.col("model").map_elements(lambda x: x.mean()).alias("mean"),
            pl.lit(contig).alias("contig")
        ]).drop("model")
    log.debug(f"{contig}-{mod_type}, done")
    return result

def worker_function(args, counter, lock, sequence, motifs, log_dir, verbose):
    """
    Process a single subpileup for one contig and one modtype

    Parameters:
    - args (tuple): The arguments to the function: contig, modtype, subpileup
    - counter (multiprocessing.Value): The progress counter
    - lock (multiprocessing.Lock): The lock for the progress counter
    - assembly (Assembly): The assembly to be processed.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - padding (int): The padding to use for the motif.
    """
    warnings.filterwarnings("ignore")
    contig, modtype, subpileup = args
    
    process_id = os.getpid()
    log_file = f"{log_dir}/score-motifs.{process_id}.log"
    configure_logger(log_file, verbose=verbose)

    try:
        result = score_contig_with_motifs(contig, modtype, subpileup, sequence, motifs)
        with lock:
            counter.value += 1
        return result
    except:
        with lock:
            counter.value += 1
        return None

def score_sample_parallel(
        assembly, pileup, motifs,
        threads = 2,
        threshold_methylation_general = 0.5,
        threshold_valid_coverage = 1,
        verbose = False,
        log_dir = None
    ):
    """
    Process a single sample
    
    Parameters:
    - assembly (Assembly): The assembly to be processed.
    - pileup (Pileup): The pileup to be processed.
    - max_candidate_size (int): The maximum size of the candidate motifs.
    - min_read_methylation_fraction (float): The minimum fraction of reads that must be methylated for a position to be considered methylated.
    - threshold_valid_coverage (int): The minimum number of reads that must cover a position for it to be considered valid.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - min_cdf_score (float): Minimum score of 1 - cdf(cdf_position) for a motif to be considered valid.
    - cdf_position (float): The position to evaluate the cdf at.
    - min_motif_frequency (int): Used to get minimum number of sequences to evaluate motif at.
    """
    assert pileup is not None, "Pileup is None"
    assert assembly is not None, "Assembly is None"
    assert threshold_methylation_general >= 0 and threshold_methylation_general <= 1, "threshold_methylation_general must be between 0 and 1"
    assert threshold_valid_coverage >= 0, "min_valid_coverage must be greater than 0"

    # Filter pileup
    pileup = pileup \
            .filter(pl.col("Nvalid_cov") > threshold_valid_coverage) \
            .filter(pl.col("fraction_mod") >= threshold_methylation_general) \
            .sort("contig") 
    
    # Create a list of tasks (TODO: not have a list of all data)
    tasks = [(contig, modtype, subpileup) for (contig, modtype), subpileup in pileup.groupby(["contig", "mod_type"])]

    # Create a progress manager
    manager = multiprocessing.Manager()
    counter = manager.Value('i', 0)
    lock = manager.Lock()

    # Create a pool of workers
    pool = get_context("spawn").Pool(processes=threads)

    # Create a process for the progress bar
    progress_bar_process = multiprocessing.Process(target=update_progress_bar, args=(counter, len(tasks)))
    progress_bar_process.start()

    # Put them workers to work
    results = pool.starmap(worker_function, [(
        task, 
        counter, lock, 
        assembly.assembly[task[0]].sequence, motifs,
        log_dir, verbose
        ) for task in tasks])
    results = [result for result in results if result is not None]

    # Close the pool
    pool.close()
    pool.join()

    # Close the progress bar
    progress_bar_process.join()

    if len(results) == 0:
        return None
    motifs_all_scored = pl.concat(results)
    return motifs_all_scored



def score_all_contigs_all_motifs(pileup, assm, motifs, frac_mod=0.7):
    # Score all contigs with identified motifs
    all_motifs_scored = {
        "contig":[],
        "type":[],
        "motif":[],
        "mod_position":[],
        "model":[],
        "mean":[]
    }
    for contig in pileup.pileup.get_column("contig").unique().to_list():
        for row in motifs.select("motif", "mod_position", "mod_type").unique().rows():
            model = evaluate.motif_model_contig(
                pileup.pileup.filter(pl.col("Nvalid_cov") > 1).filter(pl.col("fraction_mod") > frac_mod),
                assm.assembly[contig].sequence, 
                candidate.Motif(row[0], row[1])
            )
            all_motifs_scored["contig"].append(contig)
            all_motifs_scored["motif"].append(row[0])
            all_motifs_scored["model"].append(model)
            all_motifs_scored["mean"].append(model.mean())
            all_motifs_scored["type"].append(row[2])
            all_motifs_scored["mod_position"].append(row[1])

    scored_all = pl.DataFrame(all_motifs_scored).with_columns([
        pl.col("model").map_elements(lambda x: x._alpha).alias("alpha"),
        pl.col("model").map_elements(lambda x: x._beta).alias("beta")
    ])
    return scored_all
